@@ -1,27 +1,49 @@
-# python-mpd: Python MPD client library
+# python-mpd2: Python MPD client library
 # Copyright (C) 2008-2010  J. Alexander Treuman <jat@spatialrift.net>
+# Copyright (C) 2012  J. Thalheim <jthalheim@gmail.com>
 #
-# python-mpd is free software: you can redistribute it and/or modify
+# python-mpd2 is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
 # the Free Software Foundation, either version 3 of the License, or
 # (at your option) any later version.
 #
-# python-mpd is distributed in the hope that it will be useful,
+# python-mpd2 is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU Lesser General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public License
-# along with python-mpd.  If not, see <http://www.gnu.org/licenses/>.
+# along with python-mpd2.  If not, see <http://www.gnu.org/licenses/>.
 
+import logging
+import sys
 import socket
+import warnings
+from collections import Callable
 
-
+VERSION = (0, 5, 3)
 HELLO_PREFIX = "OK MPD "
 ERROR_PREFIX = "ACK "
 SUCCESS = "OK"
 NEXT = "list_OK"
 
+IS_PYTHON2 = sys.version_info < (3, 0)
+if IS_PYTHON2:
+    decode_str = lambda s: s.decode("utf-8")
+    encode_str = lambda s: s if type(s) == str else (unicode(s)).encode("utf-8")
+else:
+    decode_str = lambda s: s
+    encode_str = lambda s: str(s)
+
+try:
+    from logging import NullHandler
+except ImportError: # NullHandler was introduced in python2.7
+    class NullHandler(logging.Handler):
+        def emit(self, record):
+            pass
+
+logger = logging.getLogger(__name__)
+logger.addHandler(NullHandler())
 
 class MPDError(Exception):
     pass
@@ -52,164 +74,158 @@ class _NotConnected(object):
     def _dummy(*args):
         raise ConnectionError("Not connected")
 
+_commands = {
+    # Status Commands
+    "clearerror":         "_fetch_nothing",
+    "currentsong":        "_fetch_object",
+    "idle":               "_fetch_idle",
+    "status":             "_fetch_object",
+    "stats":              "_fetch_object",
+    # Playback Option Commands
+    "consume":            "_fetch_nothing",
+    "crossfade":          "_fetch_nothing",
+    "mixrampdb":          "_fetch_nothing",
+    "mixrampdelay":       "_fetch_nothing",
+    "random":             "_fetch_nothing",
+    "repeat":             "_fetch_nothing",
+    "setvol":             "_fetch_nothing",
+    "single":             "_fetch_nothing",
+    "replay_gain_mode":   "_fetch_nothing",
+    "replay_gain_status": "_fetch_item",
+    # Playback Control Commands
+    "next":               "_fetch_nothing",
+    "pause":              "_fetch_nothing",
+    "play":               "_fetch_nothing",
+    "playid":             "_fetch_nothing",
+    "previous":           "_fetch_nothing",
+    "seek":               "_fetch_nothing",
+    "seekid":             "_fetch_nothing",
+    "seekcur":            "_fetch_nothing",
+    "stop":               "_fetch_nothing",
+    # Playlist Commands
+    "add":                "_fetch_nothing",
+    "addid":              "_fetch_item",
+    "clear":              "_fetch_nothing",
+    "delete":             "_fetch_nothing",
+    "deleteid":           "_fetch_nothing",
+    "move":               "_fetch_nothing",
+    "moveid":             "_fetch_nothing",
+    "playlist":           "_fetch_playlist",
+    "playlistfind":       "_fetch_songs",
+    "playlistid":         "_fetch_songs",
+    "playlistinfo":       "_fetch_songs",
+    "playlistsearch":     "_fetch_songs",
+    "plchanges":          "_fetch_songs",
+    "plchangesposid":     "_fetch_changes",
+    "prio":               "_fetch_nothing",
+    "prioid":             "_fetch_nothing",
+    "shuffle":            "_fetch_nothing",
+    "swap":               "_fetch_nothing",
+    "swapid":             "_fetch_nothing",
+    # Stored Playlist Commands
+    "listplaylist":       "_fetch_list",
+    "listplaylistinfo":   "_fetch_songs",
+    "listplaylists":      "_fetch_playlists",
+    "load":               "_fetch_nothing",
+    "playlistadd":        "_fetch_nothing",
+    "playlistclear":      "_fetch_nothing",
+    "playlistdelete":     "_fetch_nothing",
+    "playlistmove":       "_fetch_nothing",
+    "rename":             "_fetch_nothing",
+    "rm":                 "_fetch_nothing",
+    "save":               "_fetch_nothing",
+    # Database Commands
+    "count":              "_fetch_object",
+    "find":               "_fetch_songs",
+    "findadd":            "_fetch_nothing",
+    "list":               "_fetch_list",
+    "listall":            "_fetch_database",
+    "listallinfo":        "_fetch_database",
+    "lsinfo":             "_fetch_database",
+    "readcomments":       "_fetch_object",
+    "search":             "_fetch_songs",
+    "searchadd":          "_fetch_nothing",
+    "searchaddpl":        "_fetch_nothing",
+    "update":             "_fetch_item",
+    "rescan":             "_fetch_item",
+    # Sticker Commands
+    "sticker get":        "_fetch_sticker",
+    "sticker set":        "_fetch_nothing",
+    "sticker delete":     "_fetch_nothing",
+    "sticker list":       "_fetch_stickers",
+    "sticker find":       "_fetch_songs",
+    # Connection Commands
+    "close":              None,
+    "kill":               None,
+    "password":           "_fetch_nothing",
+    "ping":               "_fetch_nothing",
+    # Audio Output Commands
+    "disableoutput":      "_fetch_nothing",
+    "enableoutput":       "_fetch_nothing",
+    "toggleoutput":       "_fetch_nothing",
+    "outputs":            "_fetch_outputs",
+    # Reflection Commands
+    "config":             "_fetch_item",
+    "commands":           "_fetch_list",
+    "notcommands":        "_fetch_list",
+    "tagtypes":           "_fetch_list",
+    "urlhandlers":        "_fetch_list",
+    "decoders":           "_fetch_plugins",
+    # Client To Client
+    "subscribe":          "_fetch_nothing",
+    "unsubscribe":        "_fetch_nothing",
+    "channels":           "_fetch_list",
+    "readmessages":       "_fetch_messages",
+    "sendmessage":        "_fetch_nothing",
+}
+
 class MPDClient(object):
-    def __init__(self):
+    def __init__(self, use_unicode=False):
         self.iterate = False
+        self.use_unicode = use_unicode
         self._reset()
-        self._commands = {
-            # Status Commands
-            "clearerror":         self._fetch_nothing,
-            "currentsong":        self._fetch_object,
-            "idle":               self._fetch_list,
-            "noidle":             None,
-            "status":             self._fetch_object,
-            "stats":              self._fetch_object,
-            # Playback Option Commands
-            "consume":            self._fetch_nothing,
-            "crossfade":          self._fetch_nothing,
-            "mixrampdb":          self._fetch_nothing,
-            "mixrampdelay":       self._fetch_nothing,
-            "random":             self._fetch_nothing,
-            "repeat":             self._fetch_nothing,
-            "setvol":             self._fetch_nothing,
-            "single":             self._fetch_nothing,
-            "replay_gain_mode":   self._fetch_nothing,
-            "replay_gain_status": self._fetch_item,
-            "volume":             self._fetch_nothing,
-            # Playback Control Commands
-            "next":               self._fetch_nothing,
-            "pause":              self._fetch_nothing,
-            "play":               self._fetch_nothing,
-            "playid":             self._fetch_nothing,
-            "previous":           self._fetch_nothing,
-            "seek":               self._fetch_nothing,
-            "seekid":             self._fetch_nothing,
-            "stop":               self._fetch_nothing,
-            # Playlist Commands
-            "add":                self._fetch_nothing,
-            "addid":              self._fetch_item,
-            "clear":              self._fetch_nothing,
-            "delete":             self._fetch_nothing,
-            "deleteid":           self._fetch_nothing,
-            "move":               self._fetch_nothing,
-            "moveid":             self._fetch_nothing,
-            "playlist":           self._fetch_playlist,
-            "playlistfind":       self._fetch_songs,
-            "playlistid":         self._fetch_songs,
-            "playlistinfo":       self._fetch_songs,
-            "playlistsearch":     self._fetch_songs,
-            "plchanges":          self._fetch_songs,
-            "plchangesposid":     self._fetch_changes,
-            "shuffle":            self._fetch_nothing,
-            "swap":               self._fetch_nothing,
-            "swapid":             self._fetch_nothing,
-            # Stored Playlist Commands
-            "listplaylist":       self._fetch_list,
-            "listplaylistinfo":   self._fetch_songs,
-            "listplaylists":      self._fetch_playlists,
-            "load":               self._fetch_nothing,
-            "playlistadd":        self._fetch_nothing,
-            "playlistclear":      self._fetch_nothing,
-            "playlistdelete":     self._fetch_nothing,
-            "playlistmove":       self._fetch_nothing,
-            "rename":             self._fetch_nothing,
-            "rm":                 self._fetch_nothing,
-            "save":               self._fetch_nothing,
-            # Database Commands
-            "count":              self._fetch_object,
-            "find":               self._fetch_songs,
-            "findadd":            self._fetch_nothing,
-            "list":               self._fetch_list,
-            "listall":            self._fetch_database,
-            "listallinfo":        self._fetch_database,
-            "lsinfo":             self._fetch_database,
-            "search":             self._fetch_songs,
-            "update":             self._fetch_item,
-            "rescan":             self._fetch_item,
-            # Sticker Commands
-            "sticker get":        self._fetch_item,
-            "sticker set":        self._fetch_nothing,
-            "sticker delete":     self._fetch_nothing,
-            "sticker list":       self._fetch_list,
-            "sticker find":       self._fetch_songs,
-            # Connection Commands
-            "close":              None,
-            "kill":               None,
-            "password":           self._fetch_nothing,
-            "ping":               self._fetch_nothing,
-            # Audio Output Commands
-            "disableoutput":      self._fetch_nothing,
-            "enableoutput":       self._fetch_nothing,
-            "outputs":            self._fetch_outputs,
-            # Reflection Commands
-            "commands":           self._fetch_list,
-            "notcommands":        self._fetch_list,
-            "tagtypes":           self._fetch_list,
-            "urlhandlers":        self._fetch_list,
-            "decoders":           self._fetch_plugins,
-        }
 
-    def __getattr__(self, attr):
-        if attr.startswith("send_"):
-            command = attr.replace("send_", "", 1)
-            wrapper = self._send
-        elif attr.startswith("fetch_"):
-            command = attr.replace("fetch_", "", 1)
-            wrapper = self._fetch
-        else:
-            command = attr
-            wrapper = self._execute
-        if command not in self._commands:
-            command = command.replace("_", " ")
-            if command not in self._commands:
-                raise AttributeError("'%s' object has no attribute '%s'" %
-                                     (self.__class__.__name__, attr))
-        return lambda *args: wrapper(command, args)
-
-    def _send(self, command, args):
+    def _send(self, command, args, retval):
         if self._command_list is not None:
             raise CommandListError("Cannot use send_%s in a command list" %
-                                   command.replace(" ", "_"))
+                                   command)
         self._write_command(command, args)
-        retval = self._commands[command]
         if retval is not None:
             self._pending.append(command)
 
-    def _fetch(self, command, args=None):
+    def _fetch(self, command, args, retval):
         if self._command_list is not None:
             raise CommandListError("Cannot use fetch_%s in a command list" %
-                                   command.replace(" ", "_"))
+                                   command)
         if self._iterating:
             raise IteratingError("Cannot use fetch_%s while iterating" %
-                                 command.replace(" ", "_"))
+                                 command)
         if not self._pending:
             raise PendingCommandError("No pending commands to fetch")
         if self._pending[0] != command:
             raise PendingCommandError("'%s' is not the currently "
                                       "pending command" % command)
         del self._pending[0]
-        retval = self._commands[command]
-        if callable(retval):
+        if isinstance(retval, Callable):
             return retval()
         return retval
 
-    def _execute(self, command, args):
+    def _execute(self, command, args, retval):
         if self._iterating:
             raise IteratingError("Cannot execute '%s' while iterating" %
                                  command)
         if self._pending:
             raise PendingCommandError("Cannot execute '%s' with "
                                       "pending commands" % command)
-        retval = self._commands[command]
         if self._command_list is not None:
-            if not callable(retval):
+            if not isinstance(retval, Callable):
                 raise CommandListError("'%s' not allowed in command list" %
                                         command)
             self._write_command(command, args)
             self._command_list.append(retval)
         else:
             self._write_command(command, args)
-            if callable(retval):
+            if isinstance(retval, Callable):
                 return retval()
             return retval
 
@@ -220,12 +236,27 @@ class MPDClient(object):
     def _write_command(self, command, args=[]):
         parts = [command]
         for arg in args:
-            parts.append('"%s"' % escape(str(arg)))
+            if type(arg) is tuple:
+                if len(arg) == 1:
+                    parts.append('"%d:"' % int(arg[0]))
+                else:
+                    parts.append('"%d:%d"' % (int(arg[0]), int(arg[1])))
+            else:
+                parts.append('"%s"' % escape(encode_str(arg)))
+        # Minimize logging cost if the logging is not activated.
+        if logger.isEnabledFor(logging.DEBUG):
+            if command == "password":
+                logger.debug("Calling MPD password(******)")
+            else:
+                logger.debug("Calling MPD %s%r", command, args)
         self._write_line(" ".join(parts))
 
     def _read_line(self):
         line = self._rfile.readline()
+        if self.use_unicode:
+            line = decode_str(line)
         if not line.endswith("\n"):
+            self.disconnect()
             raise ConnectionError("Connection lost while reading line")
         line = line.rstrip("\n")
         if line.startswith(ERROR_PREFIX):
@@ -295,6 +326,15 @@ class MPDClient(object):
             self._command_list = None
         self._fetch_nothing()
 
+    def _read_stickers(self):
+        for key, sticker in self._read_pairs():
+            value = sticker.split('=', 1)
+
+            if len(value) < 2:
+                raise ProtocolError("Could not parse sticker: %r" % sticker)
+
+            yield tuple(value)
+
     def _iterator_wrapper(self, iterator):
         try:
             for item in iterator:
@@ -319,6 +359,14 @@ class MPDClient(object):
             return
         return pairs[0][1]
 
+    def _fetch_sticker(self):
+        # Either we get one or we get an error while reading the line
+        key, value = list(self._read_stickers())[0]
+        return value
+
+    def _fetch_stickers(self):
+        return dict(self._read_stickers())
+
     def _fetch_list(self):
         return self._wrap_iterator(self._read_list())
 
@@ -337,6 +385,12 @@ class MPDClient(object):
     def _fetch_changes(self):
         return self._fetch_objects(["cpos"])
 
+    def _fetch_idle(self):
+        self._sock.settimeout(self.idletimeout)
+        ret = self._fetch_list()
+        self._sock.settimeout(self._timeout)
+        return ret
+
     def _fetch_songs(self):
         return self._fetch_objects(["file"])
 
@@ -346,6 +400,9 @@ class MPDClient(object):
     def _fetch_database(self):
         return self._fetch_objects(["file", "directory", "playlist"])
 
+    def _fetch_messages(self):
+        return self._fetch_objects(["channel"])
+
     def _fetch_outputs(self):
         return self._fetch_objects(["outputid"])
 
@@ -355,9 +412,17 @@ class MPDClient(object):
     def _fetch_command_list(self):
         return self._wrap_iterator(self._read_command_list())
 
+    def noidle(self):
+        if not self._pending or self._pending[0] != 'idle':
+          raise CommandError('cannot send noidle if send_idle was not called')
+        del self._pending[0]
+        self._write_command("noidle")
+        return self._fetch_list()
+
     def _hello(self):
         line = self._rfile.readline()
         if not line.endswith("\n"):
+            self.disconnect()
             raise ConnectionError("Connection lost while reading MPD hello")
         line = line.rstrip("\n")
         if not line.startswith(HELLO_PREFIX):
@@ -378,6 +443,7 @@ class MPDClient(object):
             raise ConnectionError("Unix domain sockets not supported "
                                   "on this platform")
         sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+        sock.settimeout(self.timeout)
         sock.connect(path)
         return sock
 
@@ -394,9 +460,12 @@ class MPDClient(object):
             sock = None
             try:
                 sock = socket.socket(af, socktype, proto)
+                sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
+                sock.settimeout(self.timeout)
                 sock.connect(sa)
                 return sock
-            except socket.error, err:
+            except socket.error as e:
+                err = e
                 if sock is not None:
                     sock.close()
         if err is not None:
@@ -404,15 +473,40 @@ class MPDClient(object):
         else:
             raise ConnectionError("getaddrinfo returns an empty list")
 
-    def connect(self, host, port):
+    def _settimeout(self, timeout):
+        self._timeout = timeout
+        if self._sock != None:
+            self._sock.settimeout(timeout)
+    def _gettimeout(self):
+        return self._timeout
+    timeout = property(_gettimeout, _settimeout)
+    _timeout = None
+    idletimeout = None
+
+    def connect(self, host, port, timeout=None):
+        logger.info("Calling MPD connect(%r, %r, timeout=%r)", host,
+                     port, timeout)
         if self._sock is not None:
             raise ConnectionError("Already connected")
+        if timeout != None:
+            warnings.warn("The timeout parameter in connect() is deprecated! "
+                          "Use MPDClient.timeout = yourtimeout instead.",
+                          DeprecationWarning)
+            self.timeout = timeout
         if host.startswith("/"):
             self._sock = self._connect_unix(host)
         else:
             self._sock = self._connect_tcp(host, port)
-        self._rfile = self._sock.makefile("rb")
-        self._wfile = self._sock.makefile("wb")
+
+        if IS_PYTHON2:
+            self._rfile = self._sock.makefile("r")
+            self._wfile = self._sock.makefile("w")
+        else:
+            # Force UTF-8 encoding, since this is dependant from the LC_CTYPE
+            # locale.
+            self._rfile = self._sock.makefile("r", encoding="utf-8")
+            self._wfile = self._sock.makefile("w", encoding="utf-8")
+
         try:
             self._hello()
         except:
@@ -420,9 +514,13 @@ class MPDClient(object):
             raise
 
     def disconnect(self):
-        self._rfile.close()
-        self._wfile.close()
-        self._sock.close()
+        logger.info("Calling MPD disconnect()")
+        if not self._rfile is None:
+            self._rfile.close()
+        if not self._wfile is None:
+            self._wfile.close()
+        if not self._sock is None:
+            self._sock.close()
         self._reset()
 
     def fileno(self):
@@ -449,6 +547,45 @@ class MPDClient(object):
         self._write_command("command_list_end")
         return self._fetch_command_list()
 
+    @classmethod
+    def add_command(cls, name, callback):
+        method = newFunction(cls._execute, key, callback)
+        send_method = newFunction(cls._send, key, callback)
+        fetch_method = newFunction(cls._fetch, key, callback)
+
+        # create new mpd commands as function in three flavors:
+        # normal, with "send_"-prefix and with "fetch_"-prefix
+        escaped_name = name.replace(" ", "_")
+        setattr(cls, escaped_name, method)
+        setattr(cls, "send_"+escaped_name, send_method)
+        setattr(cls, "fetch_"+escaped_name, fetch_method)
+
+    @classmethod
+    def remove_command(cls, name):
+        if not hasattr(cls, name):
+            raise ValueError("Can't remove not existent '%s' command" % name)
+        name = name.replace(" ", "_")
+        delattr(cls, str(name))
+        delattr(cls, str("send_" + name))
+        delattr(cls, str("fetch_" + name))
+
+def bound_decorator(self, function):
+    """ bind decorator to self """
+    if not isinstance(function, Callable):
+        return None
+
+    def decorator(*args, **kwargs):
+        return function(self, *args, **kwargs)
+    return decorator
+
+def newFunction(wrapper, name, returnValue):
+    def decorator(self, *args):
+        return wrapper(self, name, args, bound_decorator(self, returnValue))
+    return decorator
+
+for key, value in _commands.items():
+    returnValue = None if value is None else MPDClient.__dict__[value]
+    MPDClient.add_command(key, returnValue)
 
 def escape(text):
     return text.replace("\\", "\\\\").replace('"', '\\"')
